/**
 * @class
 * @constructor
 * @param {'ocx' | 'websocket'} mode - 设置webAgent mode
 */
function WebAgent(mode) {
  this.mode = mode;
  this.agent = null;
  this.webSocketConfig = null;
  this.cacheMessageQueue = [];
  this.agentId = null;
  this.guestId = null;

  /**
   * @desc 初始化
   * @return Promise
   */
  this.init = function () {
    Logger.info('init');
    var mode = this.mode, promise;
    typeIsMultiAndThrow([
      [mode, TYPE.String]
    ]);

    switch (mode) {
      case 'ocx':
        promise = this._ocxInit();
        break;
      case 'websocket':
        promise = this._websocketInit({
          url: 'ws://localhost:8888'
        });
        break;
    }

    return promise;
  };

  /********************************************************************** 统一接口定义 ************************************************************************/
  this.showMainWindow                 = function () {
    Logger.info('showMainWindow');
    this.agent.showMainWindow();
  };
  /**
   * @desc 座席签入
   * @param {string} userId - 座席名
   * @param {string} password - 密码
   * @param {string} server - 服务地址
   * @param {string} token - token 可空
   * @param {string} tokenType - token类型 可空
   */
  this.checkin                        = function (userId, password, server, token, tokenType) {
    var busy = true;
    Logger.info('checkin.userId:%s.password:%s.server:%s.busy:%s', userId, password, server, busy);
    typeIsMultiAndThrow([
      [userId, TYPE.String],
      [password, TYPE.String],
      [server, TYPE.String],
      [busy, TYPE.Boolean]
    ]);

    this.agent.checkin(userId, password, server, busy, token, tokenType);
  };
  /**
   * @desc 签出
   */
  this.checkout                       = function () {
    Logger.info('checkout');
    this.agent.checkout();
  };
  /**
   * @desc 接听
   */
  this.answer                         = function () {
    Logger.info('answer');
    this.agent.answer();
  };
  /**
   * @desc 应答加急请求
   * @param {string} serialNumber
   * @param {string} callerNumber
   * @param {boolean} agree - 是否同意加急请求
   */
   this.responseUrgent                 = function (serialNumber, callerNumber, agree) {
     Logger.info('responseUrgent.serialNumber:%s.callerNumber:%s.agree:%s', serialNumber, callerNumber, agree);
     typeIsMultiAndThrow([
       [serialNumber, TYPE.String],
       [callerNumber, TYPE.String],
       [agree, TYPE.Boolean]
     ]);

     this.agent.responseUrgent(serialNumber, callerNumber, agree);
   };
  /**
   * @desc 上报暂停
   * @param {boolean} pause - 是否暂停
   */
  this.applyPause                     = function (pause) {
    Logger.info('applyPause.pause:%s', pause);
    typeIsMultiAndThrow([
      [pause, TYPE.Boolean]
    ]);

    this.agent.applyPause(pause);
  };
  /**
   * @desc 查询指定组号的空闲座席
   * @param {string} groupId - 组号
   */
  this.queryAvailableAgents           = function (groupId) {
    Logger.info('queryAvailableAgents.groupId:%s', groupId);
    typeIsMultiAndThrow([
      [groupId, TYPE.String]
    ]);

    this.agent.queryAvailableAgents(groupId);
  };
  /**
   * @desc 转接到组
   * @param {string} groupId - 业务组id
   */
  this.transferCallToGroup            = function (groupId) {
    Logger.info('transferCallToGroup.groupId:%s', groupId);
    typeIsMultiAndThrow([
      [groupId, TYPE.String]
    ]);

    this.agent.transferCallToGroup(groupId);
  };
  /**
   * @desc 转接到指定座席
   * @param {string} userId - 座席id
   */
  this.transferCallToAgent            = function (userId) {
    Logger.info('transferCallToAgent.userId:%s', userId);
    typeIsMultiAndThrow([
      [userId, TYPE.String]
    ]);

    this.agent.transferCallToAgent(userId);
  };
  this.end                            = function () {
    Logger.info('end');
    this.agent.end()
  };
  /**
   * @desc 查询业务组号
   */
  this.queryGroups                    = function () {
    Logger.info('queryGroups');
    this.agent.queryGroups()
  };
  /**
   * @desc 开启关闭发送本地音频流
   * @param {boolean} enable - 是否开启音频流
   */
  this.enableAudio                    = function (enable) {
    Logger.info('enableAudio.enable:%s', enable);
    typeIsMultiAndThrow([
      [enable, TYPE.Boolean]
    ]);

    this.agent.enableAudio(enable);
  };
  /**
   * @desc 开启关闭发送本地视频流
   * @param {boolean} enable - 是否开启视频流
   */
  this.enableVideo                    = function (enable) {
    Logger.info('enableVideo.enable:%s', enable);
    typeIsMultiAndThrow([
      [enable, TYPE.Boolean]
    ]);

    this.agent.enableVideo(enable);
  };
  /**
   * @desc 屏幕共享
   * @param {string} enable - 是否开启屏幕共享
   */
  this.enableDesktopShare             = function (enable) {
    Logger.info('enableDesktopShare.enable:%s', enable);
    typeIsMultiAndThrow([
      [enable, TYPE.Boolean]
    ]);

    this.agent.enableDesktopShare(enable);
  };
  /**
   * @desc 实时截图
   * @param {string} videoSource - 视频源id
   */
  this.snapshot                       = function (videoSource) {
    Logger.info('snapshot.videoSource:%s', videoSource);
    typeIsMultiAndThrow([
      [videoSource, TYPE.String]
    ]);

    this.agent.snapshot(videoSource);
  };
  /**
   * @desc 实时截图访客
   */
  this.snapshotGuest                  = function () {
    Logger.info('snapshotGuest');
    this.agent.snapshotGuest();
  }

  this.sendOnlineMessage              = function (userId, content) {
    Logger.info('sendOnlineMessage.userId:%s.content:%s', userId, content);
    typeIsMultiAndThrow([
      [userId, TYPE.String],
      [content, TYPE.String]
    ]);

    this.agent.sendOnlineMessage(userId, content);
  };
   /**
   * @desc 发送会议内消息
   * @param {string} type - 消息类型，自定义，透传
   * @param {string} content - 消息内容，透传
   * @param {string} userId - 对端id，传空发给通话中所有人
   */
  this.sendConferenceMessage          = function (type, content, userId) {
    Logger.info('sendConferenceMessage.type:%s.content:%s.userId:%s', type, content, userId);
    typeIsMultiAndThrow([
      [type, TYPE.String],
      [content, TYPE.String],
      [userId, TYPE.String]
    ]);

    this.agent.sendConferenceMessage(type, content, userId);
  };
  this.getVersion                     = function () {
    Logger.info('getVersion');

    this.agent.getVersion();
  };
  /**
   * @desc 修改对端分辨率
   * @param {0 |1 | 2 | 3 | 4} resolution - 分辨率等级
   * | level | description |
   * | ----- | ----------- |
   * | 0 | 默认（自适应） |
   * | 1 | 流畅 |
   * | 2 | 普通 |
   * | 3 | 标清 |
   * | 4 | 高清 |
   */
  this.changeResolution               = function (resolution) {
    Logger.info('changeResolution.resolution:%s', resolution);
    typeIsMultiAndThrow([
      [resolution, TYPE.Number]
    ]);

    this.agent.changeResolution(resolution);
  };
  /**
   * @desc 获取可截图列表
   */
  this.getSnapshotList                = function () {
    Logger.info('getSnapshotList');

    this.agent.getSnapshotList();
  };
  /**
   * @desc 保存base64格式的图片
   * @param base64
   */
  this.saveImage                      = function (base64) {
    Logger.info('saveImage.base64.len:%s', base64.length);
    typeIsMultiAndThrow([
      [base64, TYPE.String]
    ]);

    this.agent.saveImage(base64);
  };
  /**
   * @desc 本地录制
   * @param {boolean} start - 是否开启本地录制
   * @param {number} fps - 帧率
   * @param {number} bitrate - 比特率
   * @param {number} resolutionWidth - 录制分辨率宽度
   * @param {number} resolutionHeight - 录制分辨率高度
   * @param {boolean} selfRecord - 是否录制本地画面
   */
  this.localRecord                    = function (start, fps, bitrate, resolutionWidth, resolutionHeight, selfRecord) {
    Logger.info('localRecord.start:%s.fps:%s.bitrate:%s.resolutionWidth:%s.resolutionHeight:%s.selfRecord:%s', start, fps, bitrate, resolutionWidth, resolutionHeight, selfRecord);
    typeIsMultiAndThrow([
      [start, TYPE.Boolean],
      [fps, TYPE.Number, TYPE.Undefined, TYPE.Null],
      [bitrate, TYPE.Number, TYPE.Undefined, TYPE.Null],
      [resolutionWidth, TYPE.Number, TYPE.Undefined, TYPE.Null],
      [resolutionHeight, TYPE.Number, TYPE.Undefined, TYPE.Null],
      [selfRecord, TYPE.Boolean, TYPE.Undefined, TYPE.Null],
    ]);

    this.agent.localRecord(start, fps, bitrate, resolutionWidth, resolutionHeight, selfRecord);
  };
  /**
   * @desc 邀请座席
   * @param userId - 座席id
   */
  this.invite                         = function (userId) {
    Logger.info('invite.userId:%s', userId);
    typeIsMultiAndThrow([
      [userId, TYPE.String]
    ]);

    this.agent.invite(userId);
  };
  this.destroy                        = function () {
    Logger.info('destroy');

    this.agent.destroy();
  };
  this.watchNetWorkSpeed              = function (start) {
    Logger.info('watchNetWorkSpeed.start:%s', start);
    typeIsMultiAndThrow([
      [start, TYPE.Boolean]
    ]);

    this.agent.watchNetWorkSpeed(start);
  }
  /**
   * @desc 设置appKey
   * @param appKey
   */
  this.setAppKey                      = function (appKey) {
    Logger.info('send appKey:%s', appKey);
    typeIsMultiAndThrow([
      [appKey, TYPE.String]
    ]);

    this.agent.setAppKey(appKey);
  }
  /**
   * @desc 打开检测窗口
   */
  this.showDeviceWindow                = function () {
    Logger.info('showDeviceWindow');
    this.agent.showDeviceWindow();
  }
  /**
   * @desc 设置窗口模式
   * @param type - 类型，0窗口，1控件
   * @param x - x坐标
   * @param y - y坐标
   * @param width - 宽度
   * @param height - 高度
   */
  this.setMainWindowFrame                = function (type, x, y, width, height) {
    Logger.info('setMainWindowFrame.type:%s.x:%s.y:%s.width:%s.height:%s', type, x, y, width, height);
    typeIsMultiAndThrow([
      [type, TYPE.Number],
      [x, TYPE.Number],
      [y, TYPE.Number],
      [width, TYPE.Number],
      [height, TYPE.Number]
    ]);
    this.agent.setMainWindowFrame(type, x, y, width, height);
  }
  /**
   * @desc 设置插件是否隐藏接听界面，需要在来电前设置
   * @param {boolean} - 是否隐藏
   */
  this.hideAnswerWindow                 = function (hide) {
    Logger.info('hideAnswerWindow.hide:%s', hide);
    typeIsMultiAndThrow([
      [hide, TYPE.Boolean]
    ]);
    this.agent.hideAnswerWindow(hide);
  }
  /**
   * @desc 邀请远程控制
   * @param {string} - 邀请对象id，传空默认邀请访客
   */
  this.inviteRemoteDesktop              = function (userId) {
    Logger.info('inviteRemoteDesktop.userId:%s', userId);
    this.agent.inviteRemoteDesktop(userId)
  }
  /**
   * @desc 获取访客id
   */
  this.getGuestId                       = function () {
    return this.guestId
  }
  /**
   * @desc 获取座席id
   */
  this.getAgentId                       = function (){
    return this.agentId
  }
  /**
   * @desc 设置插件的主题颜色
   * @param {string} color - 主题颜色，默认#0072ff
   */
  this.setThemeColor              = function (color) {
    Logger.info('setThemeColor.color:%s', color);
    this.agent.setThemeColor(color)
  }
  /**
   * @desc 会话保持
   * @param {boolean} hold - 是否保持
   */
  this.hold              = function (hold) {
    Logger.info('hold.hold:%s', hold);
    this.agent.hold(hold)
  }
  /**
   * @desc 发送消息
   * @param {string} userId - 用户id
   * @param {string} message - 消息
   */
    this.sendTextMessage = function (userId, message) {
    Logger.info('sendTextMessage.userId:%s.message:%s', userId, message);
    this.agent.sendTextMessage(userId, message)
  }
  /**
   * @desc 电子签名
   * @param {string} userId - 用户id
   */
  this.sign = function (userId) {
    Logger.info('sign.userId:%s', userId);
    this.agent.sign(userId)
  }
  /**
   * @desc 语音视频切换
   * @param {number} callType - 音视频类型，0是音频，1是视频
   */
  this.turnCallType = function (callType) {
    Logger.info('turnCallType.callType:%s', callType);
    this.agent.turnCallType(callType)
  }
  /**
   * @desc 静音
   * @param {boolean} mute - 是否静音
   */
  this.mute = function (mute) {
    Logger.info('mute.mute:%s', mute);
    this.agent.mute(mute)
  }

  /**
   * @desc 风险揭示
   * @param {boolean} open - 是否打开
   */
  this.risk = function (open) {
    Logger.info('risk.open:%s', open);
    this.agent.risk(open)
  }

  /********************************************************************** 统一回调事件定义 ************************************************************************/
  /**
   * @desc 检出检入状态状态改变
   * @callback oncheckstatechange
   * @param {number} state - 检出检入状态
   * | state | description |
   * | ----- | ----------- |
   * | 0 | checkout |
   * | 1 | checking |
   * | 2 | checked |
   */
  this.oncheckstatechange             = function (state) { console.log('oncheckstatechange', state); };
  /**
   * @desc 呼叫状态改变
   * @callback oncallstatechange
   * @param {number} type - 呼叫状态
   * | type | description |
   * | ----- | ----------- |
   * | 0 | incoming |
   * | 1 | talking |
   * | 2 | termed |
   * @param {string} extraString - 随路参数
   * @param {number} reason - 状态改变原因
   * @param {number} serialNumber - callid
   */
  this.oncallstatechange              = function (type, extraString, reason, serialNumber) { console.log('oncallstatechange', type, extraString, reason, serialNumber); };
  /**
   * @desc 接收到加急请求
   * @callback onurgentrequestrecvnotify
   * @param {string} srialNumber
   * @param {string} callerNumber
   */
  this.onurgentrequestrecvnotify    = function (srialNumber, callerNumber) { console.log('onurgentrecvnotify', srialNumber, callerNumber); };
  /**
   * @desc 示忙/示闲回调
   * @callback onapplyresult
   * @param applyId
   * @param {boolean} pause - 是否是示忙
   * @param {boolean} succ - 接口调用是否成功
   */
  this.onapplyresult                  = function (applyId, pause, succ) { console.log('onapplyresult', applyId, pause, succ); };
  /**
   * @desc 查询空闲座席回调
   * @callback onqueryavailableagentsresult
   * @param {boolean} succ - 查询是否成功
   * @param {Array} info - 空闲座席列表
   */
  this.onqueryavailableagentsresult   = function (succ, info) { console.log('onqueryavailableagentsresult', succ, info); };
  /**
   * @desc 查询业务组回调
   * @callback onquerygroupsresult
   * @param {boolean} succ - 查询是否成功
   * @param {Array} info - 业务组列表
   */
  this.onquerygroupsresult            = function (succ, info) { console.log('onquerygroupsresult', succ, info); };
  /**
   * @desc 实施截图结果回调
   * @callback onsnapshotresult
   * @param {boolean} succ - 是否成功
   * @param {string} base64 - base64字符串格式的图片
   * @param {string} fileName - 文件名
   */
  this.onsnapshotresult               = function (succ, base64, fileName) { console.log('onsnapshotresult', succ, base64, fileName); };
  /**
   * @desc 收到 online message 的回调
   * @callback ononlinemessagerecvnotify
   * @param {string} content - 消息内容
   * @param {string} userId - 发送者id
   */
  this.ononlinemessagerecvnotify      = function (content, userId) { console.log('ononlinemessagerecvnotify', content, userId); };
  /**
   * @desc online message 消息发送是否成功的回调
   * @callback ononlinemessagesendresult
   * @param {boolean} succ - 是否成功
   */
  this.ononlinemessagesendresult      = function (succ) { console.log('ononlinemessagesendresult', succ); };
  /**
   * @desc 收到会议中消息的回调
   * @callback onconferencemessagerecvnotify
   * @param {string} type - 消息类型
   * @param {string} content - 消息内容
   * @param {string} userId - 发送者id
   */
  this.onconferencemessagerecvnotify  = function (type, content, userId) { console.log('onconferencemessagerecvnotify', type, content, userId); };
  /**
   * @desc 会议消息发送是否成功的回调
   * @callback onconferencemessagesendresult
   * @param {boolean} result - 是否成功
   */
  this.onconferencemessagesendresult  = function (result) { console.log('onconferencemessagesendresult', result); };
  /**
   * @desc 排队人数变化回调
   * @callback onwaitcountnotify
   * @param {number} count - 当前排队人数
   */
  this.onwaitcountnotify              = function (count) { console.log('onwaitcountnotify', count); };
  /**
   * @desc 获取版本号回调
   * @callback ongetversionresult
   * @param {string} version - 版本号
   */
  this.ongetversionresult             = function (version) { console.log('ongetversionresult', version); };
  /**
   * @desc 获取可截图视频源列表
   * @callback ongetsnapshotlistresult
   * @param {Array} list - 视频源列表
   */
  this.ongetsnapshotlistresult        = function (list) { console.log('ongetsnapshotlistresult', list); };
  /**
   * @desc 本地录制回调
   * @callback onlocalrecordnotify
   * @param {boolean} isRecording - 是否正在录制
   */
  this.onlocalrecordnotify            = function (isRecording) { console.log('onlocalrecordnotify', isRecording); };
  /**
   * @desc 转接结果回调
   * @callback ontransferresult
   * @param {boolean} result - 转接是否成功
   */
  this.ontransferresult               = function (result) { console.log('ontransferresult', result); };
  /**
   * @desc 屏幕共享状态
   * @callback onscreenshareresult
   * @param {boolean} start - 屏幕共享是否开启
   */
  this.onscreenshareresult            = function (start, shareUserId) { console.log('onscreenshareresult', start, 'shareUserId', shareUserId); };
  /**
   * @desc 上下行流量事件
   * @callback onscreenshareresult
   * @param {number} uploadBytePerSecond - 上行流量
   * @param {number} downloadBytePerSecond - 下行流量
   */
  this.onwatchnetworkspeednotify      = function (uploadBytePerSecond, downloadBytePerSecond) { console.log('onscreenshareresult', uploadBytePerSecond, downloadBytePerSecond); };
  /**
   * @desc 打开检测窗口回调
   * @callback ondevicecheckresult
   * @param {boolean} result - 检测结果
   */
  this.ondevicecheckresult            = function (result) { console.log('ondevicecheckresult', result); };

  /**
   * @desc 成员网络状态变化上报
   * @callback onnetworkchangednotify
   * @param {number} network - 网络状态: 0 断开连接, 1 非常差, 2 差, 3 普通, 4 好, 5 非常好
   * @param {string} userId - 该网络状态对应用户名
   */
  this.onnetworkchangednotify         = function (network, userId) { console.log('onnetworkchangednotify', network, userId)}

  /**
   * @desc 收到文字消息
   * @callback ontextmessagerecvnotify
   * @param {string} userId - 用户id
   * @param {string} content - 收到内容
   */
  this.ontextmessagerecvnotify = function (userId, content) { console.log('ontextmessagerecvnotify', userId, content)}

  /**
   * @desc 语音视频切换结果
   * @callback oncalltypeturnednotify
   * @param {number} callType - 音视频类型，0表示音频，1表示视频
   */
  this.oncalltypeturnednotify = function (callType) { console.log('oncalltypeturnednotify', callType)}

  /**
   * mode: ocx 的初始化
   * @private
   */
  this._ocxInit = function () {
    var that = this;
    function addOCXEventListener() {
      var listenerArr = [], i = 0;
      listenerArr.push(
        createOCXEventListener('OnCheckStateChange(state)', "WebAgent.logger.info('oncheckstatechange.state:%s', state); webAgent.oncheckstatechange(state)"),
        createOCXEventListener('OnCallStateChange(type, extraString, reason)', "WebAgent.logger.info('oncallstatechange.type:%s.extraString:%s.reason:%s', type, extraString, reason); webAgent.oncallstatechange(type, extraString, reason)"),
        createOCXEventListener('OnUrgentRecvNotify(srialNumber, callerNumber)', "WebAgent.logger.info('onurgentrecvnotify.srialNumber:%s.callerNumber:%s', srialNumber, callerNumber); webAgent.onurgentrecvnotify(srialNumber, callerNumber)"),
        createOCXEventListener('OnApplyResult(applyId, pause, succ)', "WebAgent.logger.info('onapplyresult.applyId:%s.pause:%s.succ:%s', applyId, pause, succ); webAgent.onapplyresult(applyId, pause, succ)"),
        createOCXEventListener('OnQueryAvailableAgentsResult(succ, info)', "WebAgent.logger.info('onqueryavailableagentsresult.succ:%s.info:%s', succ, info); webAgent.onqueryavailableagentsresult(succ, JSON.parse(info))"),
        createOCXEventListener('OnQueryGroupsResult(succ, info)', "WebAgent.logger.info('onquerygroupsresult.succ:%s.info:%s', succ, info); webAgent.onquerygroupsresult(succ, JSON.parse(info))"),
        createOCXEventListener('OnSnapshotResult(succ, base64)', "WebAgent.logger.info('onsnapshotresult.succ:%s.base64:%s', succ, base64); webAgent.onsnapshotresult(succ, base64)"),
        createOCXEventListener('OnOnlineMessageRecvNotify(content, userId)', "WebAgent.logger.info('ononlinemessagerecvnotify.content:%s.userId:%s', content, userId); webAgent.ononlinemessagerecvnotify(content, userId)"),
        createOCXEventListener('OnOnlineMessageSendResult(succ)', "WebAgent.logger.info('ononlinemessagesendresult.succ:%s', succ); webAgent.ononlinemessagesendresult(succ)"),
        createOCXEventListener('OnConferenceMessageRecvNotify(type, content, userId)', "WebAgent.logger.info('onconferencemessagerecvnotify.type:%s.content:%s.userId:%s', type, content, userId); webAgent.onconferencemessagerecvnotify(type, content, userId)"),
        createOCXEventListener('OnConferenceMessageSendResult(result)', "WebAgent.logger.info('onconferencemessagesendresult.result:%s', result); webAgent.onconferencemessagesendresult(result)"),
        createOCXEventListener('OnWaitCountNotify(count)', "WebAgent.logger.info('onwaitcountnotify.count:%s', count); webAgent.onwaitcountnotify(count)"),
        createOCXEventListener('OnGetVersionResult(version)', "WebAgent.logger.info('ongetversionresult.version:%s', version); webAgent.ongetversionresult(version)"),
        createOCXEventListener('OnGetSnapshotListResult(list)', "WebAgent.logger.info('ongetsnapshotlistresult.list:%j', list); webAgent.ongetsnapshotlistresult(JSON.parse(list))"),
        createOCXEventListener('OnLocalRecordNotify(isRecording)', "WebAgent.logger.info('onlocalrecordnotify.isRecording:%s', isRecording); webAgent.onlocalrecordnotify(isRecording)"),
        createOCXEventListener('OnTransferResult(result)', "WebAgent.logger.info('ontransferresult.result:%s', result); webAgent.ontransferresult(result)"),
        createOCXEventListener('OnScreenShareResult(start,shareUserId)', "WebAgent.logger.info('onscreenshareresult.start:%s.shareUserId:%s', start, shareUserId); webAgent.onscreenshareresult(start,shareUserId)"),
        createOCXEventListener('OnWatchNetWorkSpeedNotify(uploadBytePerSecond, downloadBytePerSecond)', "WebAgent.logger.info('onwatchnetworkspeednotify.uploadBytePerSecond:%s.downloadBytePerSecond:%s', uploadBytePerSecond, downloadBytePerSecond); webAgent.onwatchnetworkspeednotify(uploadBytePerSecond, downloadBytePerSecond)"),
        createOCXEventListener('OnDeviceCheckResult(result)', "WebAgent.logger.info('ondevicecheckresult.result:%s', result); webAgent.ondevicecheckresult(result)")
      );
      for (i = listenerArr.length; i--;) {
        document.body.appendChild(listenerArr[i]);
      }
    }

    /* 注册OCX对象 */
    function registerOCXControl(reject) {
      var csax = document.createElement('object');
      csax.setAttribute("classid", "clsid:423AB4A9-2095-47D1-BF34-5D06D3025A55");
      csax.setAttribute("id", "cSharpActiveX");
      csax.setAttribute("name", "cSharpActiveX");
      csax.setAttribute("style", "display: none;");
      document.body.appendChild(csax);
      /* 判断OCX控件是否可用 */
      try {
        cSharpActiveX.GetVersion();
      } catch (e) {
        Logger.error('ocx invoke getVersion error\t%j', e);
        reject({type: 'error', error: 'NotFoundOCXError', data: e});
      }
      /* 页面卸载前销毁OCX控件 */
      function _destroyBeforeUnLoad(ev) {
        cSharpActiveX.Destroy();
        document.body.removeChild(cSharpActiveX);
        // ev.preventDefault();
        // ev.returnValue = 'asd';
        // return 'asd';
      }
      setTimeout(function () {
        try {
          window.addEventListener('beforeunload', _destroyBeforeUnLoad);
        } catch (e) {
          window.attachEvent('onbeforeunload', _destroyBeforeUnLoad);
        }
      }, 1000);
    }

    function createOCXAgent() {
      var agent = {
        showMainWindow          : function (show) { cSharpActiveX.ShowMainWindow(show); },
        checkin                 : function (userId, password, server, busy) { cSharpActiveX.CheckIn(userId, password, server, busy); },
        checkout                : function () { cSharpActiveX.Checkout(); },
        answer                  : function () { cSharpActiveX.Answer(); },
        responseUrgent          : function (serialNumber, callerNumber, agree) { cSharpActiveX.ResponseUrgent(serialNumber, callerNumber, agree); },
        applyPause              : function (pause) { cSharpActiveX.ApplyPause(pause); },
        queryAvailableAgents    : function (groupId) { cSharpActiveX.QueryAvailableAgents(groupId); },
        transferCallToGroup     : function (groupId) { cSharpActiveX.TransferCallToGroup(groupId); },
        transferCallToAgent     : function (userId) { cSharpActiveX.TransferCallToAgent(userId); },
        end                     : function () { cSharpActiveX.End(); },
        queryGroups             : function () { cSharpActiveX.QueryGroups(); },
        enableAudio             : function (enable) { cSharpActiveX.EnableAudio(enable); },
        enableVideo             : function (enable) { cSharpActiveX.EnableVideo(enable); },
        enableDesktopShare      : function (enable) { cSharpActiveX.EnableDesktopShare(enable); },
        snapshot                : function (videoSource) { cSharpActiveX.Snapshot(videoSource); },
        snapshotGuest           : function () { cSharpActiveX.SnapshotGuest(); },
        sendOnlineMessage       : function (userId, content) { cSharpActiveX.SendOnlineMessage(userId, content); },
        sendConferenceMessage   : function (type, content, userId) { cSharpActiveX.SendConferenceMessage(type, content, userId); },
        getVersion              : function () { cSharpActiveX.GetVersion(); },
        changeResolution        : function (resolution) { cSharpActiveX.ChangeResolution(resolution); },
        getSnapshotList         : function () { cSharpActiveX.GetSnapshotList(); },
        saveImage               : function (base64) { cSharpActiveX.SaveImage(base64); },
        localRecord             : function (start, fps, bitrate, resolutionWidth, resolutionHeight, selfRecord) { cSharpActiveX.LocalRecord(start, fps, bitrate, resolutionWidth, resolutionHeight, selfRecord); },
        invite                  : function (userId) { cSharpActiveX.Invite(userId); },
        watchNetWorkSpeed       : function () { cSharpActiveX.WatchNetWorkSpeed(); },
        setAppKey               : function (appKey) { cSharpActiveX.SetAppKey(appKey); },
        showDeviceWindow        : function () { cSharpActiveX.ShowDeviceWindow(); },
        setMainWindowFrame      : function (type, x, y, width, height) { cSharpActiveX.SetMainWindowFrame(type, x, y, width, height); }
        // getCameras              : function () { cSharpActiveX.GetCameras(); },
        // getAudioInputs          : function () { cSharpActiveX.GetAudioInputs(); },
        // getAudioOutputs         : function () { cSharpActiveX.GetAudioInputs(); },
        // getParticipants         : function () { cSharpActiveX.GetParticipants(); },
        // statCamera              : function (cameraId) { cSharpActiveX.StatCamera(cameraId); },
        // stopCamera              : function () { cSharpActiveX.StopCamera(); },
        // switchCamera            : function (cameraId) { cSharpActiveX.SwitchCamera(cameraId); },
        // statAudioInput          : function (inputId) { cSharpActiveX.StatAudioInput(inputId); },
        // stopAudioInput          : function (inputId) { cSharpActiveX.StopAudioInput(inputId); },
        // stopAudioInputAll       : function () { cSharpActiveX.StopAudioInputAll(); },
        // statAudioOutput         : function (outputId) { cSharpActiveX.StatAudioOutput(outputId); },
        // stopAudioOutput         : function () { cSharpActiveX.StopAudioOutput(); },
        // inviteRemoteDesktop     : function (userId) { cSharpActiveX.InviteRemoteDesktop(userId); },
        // snapshotEx              : function (videoSource) { cSharpActiveX.SnapshotEx(videoSource); },
      };

      return agent;
    }

    return new Promise(function(resolve, reject) {
      try {
        addOCXEventListener();
        registerOCXControl(reject);
        webAgent.agent = createOCXAgent();

        function beforeunload(ev) {
          that.destroy();
          ev.preventDefault();
          ev.returnValue = '确认离开';
          return '确认离开';
        }
        if (window.addEventListener) {
          window.addEventListener('beforeunload', beforeunload);
        } else {
          window.attachEvent('onbeforeunload', beforeunload);
        }
        resolve();
      } catch (e) {
        console.log(e);
        reject(e);
      }
    });
  };

  /**
   * @desc mode: websocket 的初始化
   * @param config
   * @private
   */
  this._websocketInit = function (config) {
    var that = this;
    function dispatchWebSocketEvent(message) {
      switch (message.BridgeType) {
        case 100:
          Logger.info('oncheckstatechange.state:%s.agentId:%s', message.State, message.AgentId);
          that.oncheckstatechange(message.State);
          if(message.State === 2){
            that.agentId = message.AgentId;
          }
          else if(message.State === 0){
            that.agentId = null
          }
          break;
        case 101:
          Logger.info('oncallstatechange.type:%s.extraString:%s.reason:%s.serialNumber:%s.guestId:%s', message.Type, message.ExtraString, message.Reason, message.SerialNumber, message.GuestId);
          that.oncallstatechange(message.Type, message.ExtraString, message.Reason, message.SerialNumber);
          if(message.Type === 1){
            that.guestId = message.GuestId
          }else{
            that.guestId = null
          }
          break;
        case 102:
          Logger.info('onurgentrequestrecvnotify.serialNumber:%s.userId:%s', message.SerialNumber, message.UserId)
          that.onurgentrequestrecvnotify(message.SerialNumber, message.UserId)
          break;
        case 103:
          Logger.info('onapplyresult.pause:%s.succ:%s', message.Pause, message.Succ);
          that.onapplyresult(null, message.Pause, message.Succ);
          break;
        case 104:
          message.Agents = JSON.parse(message.Agents);
          Logger.info('onqueryavailableagentsresult.succ:%s.info:%s', message.Succ, message.Agents);
          that.onqueryavailableagentsresult(message.Succ, message.Agents);
          break;
        case 105:
          Logger.info('onquerygroupsresult.succ:%s.groups:%j', message.Result, message.Groups);
          that.onquerygroupsresult(message.Result, message.Groups);
          break;
        case 106:
          Logger.info('onsnapshotresult.succ:%s.base64:too long!.fileName:%s', message.Succ, message.FileName);
          that.onsnapshotresult(message.Succ, message.Content, message.FileName);
          break;
        case 107:
          Logger.info('ononlinemessagerecvnotify.content:%s.userId:%s', message.Content, message.UserId);
          that.ononlinemessagerecvnotify(message.Content, message.UserId);
          break;
        case 108:
          Logger.info('ononlinemessagesendresult.result:%s', message.Result);
          that.ononlinemessagesendresult(message.Result);
          break;
        case 109:
          Logger.info('onconferencemessagerecvnotify.type:%s.content:%s.userId:%s', message.Type, message.Content, message.UserId);
          that.onconferencemessagerecvnotify(message.Type, message.Content, message.UserId);
          break;
        case 110:
          Logger.info('onconferencemessagesendresult.result:%s', message.Result);
          that.onconferencemessagesendresult(message.Result);
          break;
        case 111:
          Logger.info('onwaitcountnotify.count:%s', message.Count);
          that.onwaitcountnotify(message.Count);
          break;
        case 112:
          Logger.info('ongetversionresult.version:%s', message.Version);
          that.ongetversionresult(message.Version);
          break;
        case 113:
          message.SnapshotList = JSON.parse(message.SnapshotList);
          Logger.info('ongetsnapshotlistresult.snapshotList:%s', message.SnapshotList);
          that.ongetsnapshotlistresult(message.SnapshotList);
          break;
        case 114:
          Logger.info('onlocalrecordnotify.isRecording:%s', message.IsRecording);
          that.onlocalrecordnotify(message.IsRecording);
          break;
        case 115:
          Logger.info('ontransferresult.result:%s', message.Result);
          that.ontransferresult(message.Result);
          break;
        case 116:
          Logger.info('onscreenshareresult.start:%s', message.Start,'onscreenshareresult.shareUserId:%s',message.ShareUserId);
          that.onscreenshareresult(message.Start,message.ShareUserId);
          break;
        case 117:
          that.onwatchnetworkspeednotify(message.UploadBytePerSecond, message.DownloadBytePerSecond);
          break;
        case 118:
          Logger.info('ondevicecheckresult.result:%s', message.Result);
          that.ondevicecheckresult(message.Result);
          break;
        case 119:
          Logger.info('onnetworkchangednotify.network:%s.userId:%s', message.Network, message.UserId)
          that.onnetworkchangednotify(message.Network, message.UserId)
          break;
        case 120:
          Logger.info('ontextmessagerecvnotify.userId:%s.content:%s', message.UserId, message.Content)
          that.ontextmessagerecvnotify(message.UserId, message.Content)
          break;
        case 121:
          Logger.info('oncalltypeturnednotify.callType:%s', message.CallType)
          that.oncalltypeturnednotify(message.CallType)
          break;
      }
    }

    function connectWebSocket(config) {
      if (config === null || config === undefined ||
        config.url === null || config.url === undefined) {
        throw new Error('connect webSocket config error: ' + JSON.stringify(config));
      }

      var maxRetry = 10;
      function _retryConnect(resolve, reject) {
        var ws, timeoutKey = -1;
        ws = new WebSocket(config.url);
        ws.onopen = function (ev) {
          Logger.info('websocket open\twebsocket address: %s\tremaining reconnect times: %d', config.url, maxRetry);
          maxRetry = 5;
          var cacheMessageQueue = that.cacheMessageQueue;
          that.ws = ws;
          while (cacheMessageQueue.length > 0) {
            ws.send(JSON.stringify(cacheMessageQueue.shift()));
          }
          resolve();
        };
        ws.onerror = function (ev) { Logger.error('websocket error.\t%j', ev); };
        ws.onclose = function (ev) {
          Logger.info('websocket close success\tclose code: %d', ev.code);

          if (ev.code === 1000) return;
          Logger.error('websocket close [remaining reconnect time: %d]', maxRetry);
          clearTimeout(timeoutKey);
          if (maxRetry <= 0) {
            reject({error: 'ConnectError'});
          } else {
            maxRetry--;
            _retryConnect.call(this, resolve, reject);
          }
        };
        ws.onmessage = function (ev) {
          var message;
          try {
            message = JSON.parse(ev.data);
          } catch (e) {
            Logger.error('websocket message cannot parse. %j', ev.data);
            return;
          }

          dispatchWebSocketEvent(message);
        };

        timeoutKey = setTimeout(function() {
          if (ws.readyState === 1) return;
          ws.close();
        }, 5000);
      }
      return new Promise(function(resolve, reject) {
        _retryConnect(resolve, reject);
      });
    }

    function createWebSocketAgent() {
      function sendByWebSocket(message) {
        if (that.ws.readyState !== 1) {
          that.cacheMessageQueue.push(message);
        } else {
          that.ws.send(JSON.stringify(message));
        }
      }
      var agent = {
        showMainWindow          : function (show) { sendByWebSocket({BridgeType: 0, Show: show}); },
        checkin                 : function (userId, password, server, busy, token, tokenType) { sendByWebSocket({BridgeType: 1, UserId: userId, Password: password, Server: server, Busy: busy, Token: token, TokenType: tokenType}); },
        checkout                : function () { sendByWebSocket({BridgeType: 2}); },
        answer                  : function () { sendByWebSocket({BridgeType: 3}); },
        responseUrgent          : function (serialNumber, callerNumber, agree) { sendByWebSocket({BridgeType: 4, SerialNumber: serialNumber, CallerNumber: callerNumber, Agree: agree}); },
        applyPause              : function (pause) { sendByWebSocket({BridgeType: 5, Pause: pause}); },
        queryAvailableAgents    : function (groupId) { sendByWebSocket({BridgeType: 6, GroupId: groupId}); },
        transferCallToGroup     : function (groupId) { sendByWebSocket({BridgeType: 7, GroupId: groupId}); },
        transferCallToAgent     : function (userId) { sendByWebSocket({BridgeType: 8, UserId: userId}); },
        end                     : function () { sendByWebSocket({BridgeType: 9}); },
        queryGroups             : function () { sendByWebSocket({BridgeType: 10}); },
        enableAudio             : function (enable) { sendByWebSocket({BridgeType: 11, Enable: enable}); },
        enableVideo             : function (enable) { sendByWebSocket({BridgeType: 12, Enable: enable}); },
        enableDesktopShare      : function (enable) { sendByWebSocket({BridgeType: 13, Enable: enable}); },
        snapshot                : function (videoSource) { sendByWebSocket({BridgeType: 14, VideoSource: videoSource}); },
        snapshotGuest           : function () { sendByWebSocket({BridgeType: 14, ForceGuest: true}); },
        sendOnlineMessage       : function (userId, content) { sendByWebSocket({BridgeType: 15, UserId: userId, Content: content}); },
        sendConferenceMessage   : function (type, content, userId) { sendByWebSocket({BridgeType: 16, Type: type, Content: content, UserId: userId}); },
        getVersion              : function () { sendByWebSocket({BridgeType: 17}); },
        changeResolution        : function (resolution) { sendByWebSocket({BridgeType: 18, Resolution: resolution}); },
        getSnapshotList         : function () { sendByWebSocket({BridgeType: 19}); },
        saveImage               : function (base64) { sendByWebSocket({BridgeType: 20, Base64String: base64}); },
        localRecord             : function (start, fps, bitrate, resolutionWidth, resolutionHeight, selfRecord) { sendByWebSocket({BridgeType: 21, Start: start, Fps: fps, Bitrate: bitrate, ResolutionWidth: resolutionWidth, ResolutionHeight: resolutionHeight, SelfRecord: selfRecord}); },
        invite                  : function (userId) { sendByWebSocket({BridgeType: 22, UserId: userId}); },
        destroy                 : function () { sendByWebSocket({BridgeType: 23}); },
        watchNetWorkSpeed       : function (start) { sendByWebSocket({BridgeType: 24, Start: start}); },
        setAppKey               : function (appkey) { sendByWebSocket({BridgeType: 25, AppKey: appkey}); },
        showDeviceWindow        : function () { sendByWebSocket({BridgeType: 26}); },
        setMainWindowFrame      : function (type, x, y, width, height) { sendByWebSocket({BridgeType: 27, Type: type, X: x, Y: y, Width: width, Height: height}); },
        hideAnswerWindow        : function (hide) { sendByWebSocket({BridgeType: 28, Hide: hide})},
        inviteRemoteDesktop     : function (userId) { sendByWebSocket({BridgeType: 29, UserId: userId})},
        setThemeColor           : function (color) { sendByWebSocket({BridgeType: 30, Color: color})},
        hold                    : function (hold) { sendByWebSocket({BridgeType: 31, Hold: hold})},
        sendTextMessage         : function (userId, message) { sendByWebSocket({BridgeType: 32, Content: message, UserId: userId})},
        sign                    : function (userId) { sendByWebSocket({BridgeType: 33, UserId: userId})},
        turnCallType            : function (callType) { sendByWebSocket({BridgeType: 34, CallType: callType})},
        mute                    : function (mute) { sendByWebSocket({BridgeType: 35, Mute: mute})},
        risk                    : function (open) { sendByWebSocket({BridgeType: 36, Open: open})},
      };

      return agent;
    }

    function awakePlugin() {
      var iframe = document.createElement('iframe');
      iframe.setAttribute('src','Juphoon.WebAgent:');
      iframe.style.display = 'none'
      document.body.appendChild(iframe)
    }

    that.webSocketConfig = config;

    setTimeout(function () {
      awakePlugin();
    });
    return new Promise(function(resolve, reject) {
      connectWebSocket(config)
        .then(function() {
          that.agent = createWebSocketAgent();
          window.addEventListener('beforeunload', function (ev) {
            console.log('destroy');
            that.destroy();
            ev.preventDefault();
            ev.returnValue = '确认离开';
            return '确认离开';
          });
          resolve();
        })
        ['catch'](function(err) {
          reject({type: 'error', error: err.error, data: err});
        });
    });
  };

  function createOCXEventListener(event, string) {
    var listenerScript = document.createElement('script');
    listenerScript.setAttribute('language', 'javascript');
    listenerScript.setAttribute('for', 'cSharpActiveX');
    listenerScript.setAttribute('event', event);
    listenerScript.text = string;
    return listenerScript;
  }

  /********************************************************************** Logger ************************************************************************/

  var Logger = {
    LoggerLevel: {
      DEBUG: 0,
      INFO: 1,
      WARN: 2,
      ERROR: 3,
      NONE: 4,
      0: 'DEBUG',
      1: 'INFO',
      2: 'WARN',
      3: 'ERROR',
      4: 'NONE',
    },
    level: 0,
    logList: [],
    /**
     * @desc 设置日志等级
     * @param {LoggerLevel} level
     */
    setLogLevel: function(level) {
      this.level = level;
    },
    debug: function(format) {
      var args = [], i;
      args.push(this.LoggerLevel.DEBUG);
      for (i = 0; i < arguments.length; i++) args.push(arguments[i]);
      this.log.apply(this, args);
    },
    /**
     * @desc 记录{@link LoggerLevel.INFO}级别的日志
     * @param {string} format - 格式化字符串
     * @param {...*} args - 格式化参数
     */
    info: function(format) {
      var args = [], i;
      args.push(this.LoggerLevel.INFO);
      for (i = 0; i < arguments.length; i++) args.push(arguments[i]);
      this.log.apply(this, args);
    },
    /**
     * @desc 记录{@link LoggerLevel.WARN}级别的日志
     * @param {string} format - 格式化字符串
     * @param {...*} args - 格式化参数
     */
    warning: function(format) {
      var args = [], i;
      args.push(this.LoggerLevel.WARN);
      for (i = 0; i < arguments.length; i++) args.push(arguments[i]);
      this.log.apply(this, args);
    },
    /**
     * @desc 记录{@link LoggerLevel.ERROR}级别的日志
     * @param {string} format - 格式化字符串
     * @param {...*} args - 格式化参数
     */
    error: function(format) {
      var args = [], i;
      args.push(this.LoggerLevel.ERROR);
      for (i = 0; i < arguments.length; i++) args.push(arguments[i]);
      this.log.apply(this, args);
    },
    /**
     * @desc 记录指定级别的日志
     * @param {LoggerLevel} level - 日志级别
     * @param {string} format - 格式化字符串
     * @param {...*} args - 格式化参数
     */
    log: function(level, format) {
      var i, log = [new Date().toISOString() + '\t' + this.LoggerLevel[level] + '\t' + format];
      for (i = 2; i < arguments.length; i++) log.push(arguments[i]);

      this.logList.push(log);
      if (this.level <= level) { this.onlog(log); }
    },
    /**
     * @desc 以纯文本格式返回日志信息
     * @return {string}
     */
    asText: function() {
      var i, text = '';
      for (i = 0; i < this.logList.length; i++) {
        text += format.apply(this, this.logList[i]) + '\n';
      }
    },
    /**
     * @desc 以数组返回日志信息
     *
     * 数组项格式说明:
     *
     * | Index | Type | Description |
     * | ----- | ---- | ----------- |
     * | 0 | {@link number} | 日志级别 |
     * | 1 | {@link string} | 格式化字符串 |
     * | 2... | {@link ...*} | 格式化参数 |
     * @return {Array<Array<*>>}
     */
    asArray: function() {
      return this.logList;
    },
    onlog: function(logInfo) {
      console.log(format.apply(this, logInfo));
    }
  };

  WebAgent.logger = Logger;
  /********************************************************************** TypeUtils ************************************************************************/

  var TYPE = {
    Array: '[object Array]',
    Boolean: '[object Boolean]',
    BigInt: '[object BigInt]',
    Date: '[object Date]',
    Error: '[object Error]',
    Function: '[object Function]',
    Number: '[object Number]',
    Null: '[object Null]',
    Math: '[object Math]',
    MediaStream: '[object MediaStream]',
    Object: '[object Object]',
    RegExp: '[object RegExp]',
    String: '[object String]',
    Symbol: '[object Symbol]',
    Undefined: '[object Undefined]',
    UnKnown: '[object Unknown]',
  };

  var TYPE_COMPOSE = {
    UAN: [TYPE.Undefined, TYPE.Null]
  };

  function typeOf(variable) {
    var type;
    if (variable === undefined) type = TYPE.Undefined;
    else if (variable === null) type = TYPE.Null;
    else type = Object.prototype.toString.call(variable);

    return type;
  }

  function typeIs(variable) {
    if (arguments.length < 1) return false;
    var varType = typeOf(arguments[0]);
    for (var i = 1; i < arguments.length; i++)
      if (varType === arguments[i]) return true;
    return false;
  }

  function typeIsMultiAndThrow(variableParameterArr) {
    var i, error = [], variableParameter;
    if (!typeIs(variableParameterArr, TYPE.Array)) error.push('parameter not a array.');

    for (i = 0; i < variableParameterArr.length; i++) {
      variableParameter = variableParameterArr[i];

      if (!typeIs.apply(this, variableParameter)) {
        var j, temp = '';
        for (j = 1; j < variableParameter.length; j++) {
          temp += variableParameter[j];
          if (j < variableParameter.length - 1) temp += ', ';
        }
        error.push('parameter [' + variableParameter[0] + '] is not ' + temp);
      }
    }
    if (error.length > 0) throw new Error(error.join('\n'));
    return true;
  }

  function typeIsString(variable, notNull) {
    if (!notUAN(notNull)) notNull = true;
    return typeIs(variable, TYPE.String) && (notNull ? variable.length > 0 : true);
  }

  function typeIsNumber(variable, notNaN) {
    if (!notUAN(notNaN)) notNaN = true;
    return typeIs(variable, TYPE.Number) && (notNaN ? !isNaN(variable) : true);
  }

  function notUAN(variable) {
    var args = [variable], i;
    for (i = 0; i < TYPE_COMPOSE.UAN.length; i++) args.push(TYPE_COMPOSE.UAN[i]);
    return !typeIs.apply(this, args);
  }

  function isFunction(variable) {
    return typeIs(variable, TYPE.Function);
  }

  function isTrue(variable) {
    return typeIs(variable, TYPE.Boolean) && variable;
  }

  /********************************************************************** IE8语法兼容 ************************************************************************/

  if (!Date.prototype.toISOString) {
    Date.prototype.toISOString = function() {
      function pad(n) { return n < 10 ? '0' + n : n }
      return this.getUTCFullYear() + '-'
        + pad(this.getUTCMonth() + 1) + '-'
        + pad(this.getUTCDate()) + 'T'
        + pad(this.getUTCHours()) + ':'
        + pad(this.getUTCMinutes()) + ':'
        + pad(this.getUTCSeconds()) + '.'
        + pad(this.getUTCMilliseconds()) + 'Z';
    }
  }

  Logger.onlog = function (logInfo) {
    console.log(format.apply(this, logInfo));
  };
}

function commonVersionStringMaker(major, minor, patchset, status) {
  return major + '.' + minor + '.' + patchset + (status === 15 ? '' : '.' + status);
}

WebAgent.version = commonVersionStringMaker(2, 1, 1, 15);
